"""Objectives related to the bootstrap current profile."""

import numpy as np

from desc.backend import jnp
from desc.compute import get_profiles, get_transforms
from desc.compute.utils import _compute as compute_fun
from desc.grid import LinearGrid
from desc.utils import Timer, errorif, warnif

from .normalization import compute_scaling_factors
from .objective_funs import _Objective, collect_docs


class BootstrapRedlConsistency(_Objective):
    """
    Promote consistency of the bootstrap current for axisymmetry or quasi-symmetry.

    This objective function penalizes the difference between the MHD
    and neoclassical profiles of parallel current, using the Redl
    formula for the bootstrap current. The scalar objective is defined as

    f = ½ ∫dρ [(⟨J⋅B⟩_MHD - ⟨J⋅B⟩_Redl) / (J_ref B_ref)]²

    where J_ref and B_ref are the reference magnitudes of current
    density and magnetic field strength.  This objective is treated as
    a sum of Nρ least-squares terms, where Nρ is the number of rho
    grid points.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    grid : Grid, optional
        Collocation grid containing the nodes to evaluate at.
        Defaults to ``LinearGrid(eq.L_grid, eq.M_grid, eq.N_grid)``
    helicity : tuple, optional
        Type of quasi-symmetry (M, N). Default = quasi-axisymmetry (1, 0).
        First entry must be M=1. Second entry is the toroidal mode number N,
        used for evaluating the Redl bootstrap current formula. Set to 0 for axisymmetry
        or quasi-axisymmetry; set to +/-NFP for quasi-helical symmetry.

    """

    __doc__ = __doc__.rstrip() + collect_docs(
        target_default="``target=0``.", bounds_default="``target=0``."
    )

    _coordinates = "r"
    _units = "(T A m^-2)"
    _print_value_fmt = "Bootstrap current self-consistency error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        loss_function=None,
        deriv_mode="auto",
        grid=None,
        helicity=(1, 0),
        name="Bootstrap current self-consistency (Redl)",
        jac_chunk_size=None,
    ):
        if target is None and bounds is None:
            target = 0
        assert (len(helicity) == 2) and (int(helicity[1]) == helicity[1])
        assert helicity[0] == 1, "Redl bootstrap current model assumes helicity[0] == 1"
        self._grid = grid
        self.helicity = helicity
        super().__init__(
            things=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            loss_function=loss_function,
            deriv_mode=deriv_mode,
            name=name,
            jac_chunk_size=jac_chunk_size,
        )

    def build(self, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._grid is None:
            grid = LinearGrid(
                M=eq.M_grid,
                N=eq.N_grid,
                NFP=eq.NFP,
                sym=eq.sym,
                rho=np.linspace(1 / eq.L, 1, eq.L) - 1 / (2 * eq.L),
            )
        else:
            grid = self._grid

        warnif(
            (grid.num_theta * (1 + eq.sym)) < 2 * eq.M,
            RuntimeWarning,
            "BootstrapRedlConsistency objective grid requires poloidal "
            "resolution for surface averages",
        )
        warnif(
            grid.num_zeta < 2 * eq.N,
            RuntimeWarning,
            "BootstrapRedlConsistency objective grid requires toroidal "
            "resolution for surface averages",
        )

        errorif(
            not (self.helicity[1] == 0 or abs(self.helicity[1]) == eq.NFP),
            ValueError,
            "Helicity toroidal mode number should be 0 (QA) or +/- NFP (QH)",
        )

        self._dim_f = grid.num_rho
        self._data_keys = ["<J*B>", "<J*B> Redl"]

        errorif(
            eq.electron_temperature is None,
            RuntimeError,
            "Bootstrap current calculation requires an electron temperature "
            "profile.",
        )
        errorif(
            eq.electron_density is None,
            RuntimeError,
            "Bootstrap current calculation requires an electron density profile.",
        )
        errorif(
            eq.ion_temperature is None,
            RuntimeError,
            "Bootstrap current calculation requires an ion temperature profile.",
        )

        # Try to catch cases in which density or temperatures are specified in the
        # wrong units. Densities should be ~ 10^20, temperatures are ~ 10^3.
        rho = eq.compute("rho", grid=grid)["rho"]

        warnif(
            jnp.any(eq.electron_density(rho) > 1e22),
            UserWarning,
            "Electron density is surprisingly high. It should have units of "
            "1/meters^3",
        )
        warnif(
            jnp.any(eq.electron_temperature(rho) > 50e3),
            UserWarning,
            "Electron temperature is surprisingly high. It should have units of eV",
        )
        warnif(
            jnp.any(eq.ion_temperature(rho) > 50e3),
            UserWarning,
            "Ion temperature is surprisingly high. It should have units of eV",
        )
        # Profiles may go to 0 at rho=1, so exclude the last few grid points from lower
        # bounds:
        rho = rho[rho < 0.85]
        warnif(
            jnp.any(eq.electron_density(rho) < 1e17),
            UserWarning,
            "Electron density is surprisingly low. It should have units 1/meters^3",
        )
        warnif(
            jnp.any(eq.electron_temperature(rho) < 30),
            UserWarning,
            "Electron temperature is surprisingly low. It should have units of eV",
        )
        warnif(
            jnp.any(eq.ion_temperature(rho) < 30),
            UserWarning,
            "Ion temperature is surprisingly low. It should have units of eV",
        )

        timer = Timer()
        if verbose > 0:
            print("Precomputing transforms")
        timer.start("Precomputing transforms")

        profiles = get_profiles(self._data_keys, obj=eq, grid=grid)
        transforms = get_transforms(self._data_keys, obj=eq, grid=grid)

        self._constants = {
            "transforms": transforms,
            "profiles": profiles,
            "helicity": self._helicity,
        }

        timer.stop("Precomputing transforms")
        if verbose > 1:
            timer.disp("Precomputing transforms")

        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["B"] * scales["J"]

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute the bootstrap current self-consistency objective.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        obj : ndarray
            Bootstrap current self-consistency residual on each rho grid point.

        """
        if constants is None:
            constants = self.constants
        data = compute_fun(
            "desc.equilibrium.equilibrium.Equilibrium",
            self._data_keys,
            params=params,
            transforms=constants["transforms"],
            profiles=constants["profiles"],
            helicity=constants["helicity"],
        )
        return constants["transforms"]["grid"].compress(
            data["<J*B>"] - data["<J*B> Redl"]
        )

    @property
    def helicity(self):
        """tuple: Type of quasi-symmetry (M, N)."""
        return self._helicity

    @helicity.setter
    def helicity(self, helicity):
        assert (
            (len(helicity) == 2)
            and (int(helicity[0]) == helicity[0])
            and (int(helicity[1]) == helicity[1])
        )
        assert helicity[0] == 1, "Redl bootstrap current model assumes helicity[0] == 1"
        if hasattr(self, "_helicity") and self._helicity != helicity:
            self._built = False
        self._helicity = helicity



"""Tests for compute functions evaluated at limits.

If a new quantity is added to the compute functions whose limit is not finite
(or does not exist), simply add it to the ``not_finite_limits`` set below.
If the limit has yet to be derived, add it to the ``not_implemented_limits`` set.
"""

import functools
import inspect

import numpy as np
import pytest

from desc.compute import data_index
from desc.compute.utils import _grow_seeds
from desc.equilibrium import Equilibrium
from desc.examples import get
from desc.grid import LinearGrid
from desc.integrals import surface_integrals_map
from desc.objectives import GenericObjective, ObjectiveFunction
from desc.utils import dot

# Unless mentioned in the source code of the compute function, the assumptions
# made to compute the magnetic axis limit can be reduced to assuming that these
# functions tend toward zero as the magnetic axis is approached and that
# d²ψ/(dρ)² and 𝜕√𝑔/𝜕𝜌 are both finite nonzero at the magnetic axis.
# Also, dⁿψ/(dρ)ⁿ for n > 3 is assumed zero everywhere.
zero_limits = {"rho", "psi", "psi_r", "psi_rrr", "e_theta", "sqrt(g)", "B_t"}

# These compute quantities require kinetic profiles, which are not defined for all
# configurations (giving NaN values)
not_continuous_limits = {"current Redl", "P_ISS04", "P_fusion", "<sigma*nu>"}

not_finite_limits = {
    "D_Mercier",
    "D_geodesic",
    "D_well",
    "J^theta",
    "curvature_H_rho",
    "curvature_H_zeta",
    "curvature_K_rho",
    "curvature_K_zeta",
    "curvature_k1_rho",
    "curvature_k1_zeta",
    "curvature_k2_rho",
    "curvature_k2_zeta",
    "e^helical",
    "e^theta",
    "e^theta_r",
    "e^theta_t",
    "e^theta_z",
    "g^rt",
    "g^rt_r",
    "g^rt_t",
    "g^rt_z",
    "g^tt",
    "g^tt_r",
    "g^tt_t",
    "g^tt_z",
    "g^tz",
    "g^tz_r",
    "g^tz_t",
    "g^tz_z",
    "grad(alpha)",
    "g^aa",
    "g^ra",
    "gbdrift",
    "cvdrift",
    "grad(alpha)",
    "|e^helical|",
    "|grad(theta)|",
    "<J*B> Redl",  # may not exist for all configurations
}
not_implemented_limits = {
    # reliant limits will be added to this set automatically
    "D_current",
    "n_rho_z",
    "|e_theta x e_zeta|_z",
    "e^rho_rr",
    "e^theta_rr",
    "e^zeta_rr",
    "e^rho_rt",
    "e^rho_tt",
    "e^theta_rt",
    "e^theta_tt",
    "e^zeta_rt",
    "e^zeta_tt",
    "e^rho_rz",
    "e^rho_tz",
    "e^rho_zz",
    "e^theta_rz",
    "e^theta_tz",
    "e^theta_zz",
    "e^zeta_rz",
    "e^zeta_tz",
    "e^zeta_zz",
    "K_vc",  # only defined on surface
    "iota_num_rrr",
    "iota_den_rrr",
}


def add_all_aliases(names):
    """Add aliases to limits."""
    all_aliases = []
    for name in names:
        for base_class in data_index.keys():
            if name in data_index[base_class].keys():
                all_aliases.append(data_index[base_class][name]["aliases"])

    # flatten
    all_aliases = [name for sublist in all_aliases for name in sublist]
    names.update(all_aliases)

    return names


zero_limits = add_all_aliases(zero_limits)
not_finite_limits = add_all_aliases(not_finite_limits)
not_implemented_limits = add_all_aliases(not_implemented_limits)
not_implemented_limits = _grow_seeds(
    "desc.equilibrium.equilibrium.Equilibrium",
    not_implemented_limits,
    data_index["desc.equilibrium.equilibrium.Equilibrium"].keys() - not_finite_limits,
    has_axis=True,
)


def _skip_this(eq, name):
    return (
        name in not_implemented_limits
        or (eq.atomic_number is None and "Zeff" in name)
        or (eq.electron_temperature is None and "Te" in name)
        or (eq.electron_density is None and "ne" in name)
        or (eq.ion_temperature is None and "Ti" in name)
        or (eq.electron_density is None and "ni" in name)
        or (eq.anisotropy is None and "beta_a" in name)
        or (eq.pressure is not None and "<J*B> Redl" in name)
        or (eq.current is None and "iota_num" in name)
        # These quantities require a coordinate mapping to compute and special grids, so
        # it's not economical to test their axis limits here. Instead, a grid that
        # includes the axis should be used in existing unit tests for these quantities.
        or bool(
            data_index["desc.equilibrium.equilibrium.Equilibrium"][name][
                "source_grid_requirement"
            ]
        )
    )


def assert_is_continuous(
    eq,
    names=data_index["desc.equilibrium.equilibrium.Equilibrium"].keys(),
    delta=1e-4,
    rtol=1e-5,
    atol=5e-7,
    desired_at_axis=None,
    kwargs=None,
):
    """
    Asserts that the rho=0 axis limits of names are continuous extensions.

    Parameters
    ----------
    eq : Equilibrium
        The equilibrium object used for the computation.
    names : list of str
        A list of names of the quantities to test for continuity.
    delta: float, optional
        Max distance from magnetic axis.
        Smaller values accumulate finite precision error and fitting issues.
    rtol : float, optional
        Relative tolerance.
    atol : float, optional
        Absolute tolerance.
    desired_at_axis : float, optional
        If not provided, the values are extrapolated with a polynomial fit.
    kwargs : dict, optional
        Keyword arguments to override the parameters above for specific names.
        The dictionary should have the following structure:
        {
            "name1": {
                "rtol": custom_rtol1,
                "atol": custom_atol1,
                "desired_at_axis": custom_desired_at_axis1
            },
            "name2": {"rtol": custom_rtol2},
            ...
        }

    """
    if kwargs is None:
        kwargs = {}
    # TODO ( #671, #1206):currently skip Boozer quants because it need sym=False
    #      grid (#1206) and Boozer axis limits are not yet implemented (#671)
    names = [
        name
        for name in names
        if not (
            "Boozer" in name
            or "_mn" in name
            or name == "B modes"
            or _skip_this(eq, name)
        )
    ]

    num_points = 12
    rho = np.linspace(start=0, stop=delta, num=num_points)
    grid = LinearGrid(rho=rho, M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, sym=eq.sym)
    axis = grid.nodes[:, 0] == 0
    assert axis.any() and not axis.all()
    integrate = surface_integrals_map(grid, expand_out=False)
    data = eq.compute(names=names, grid=grid)

    p = "desc.equilibrium.equilibrium.Equilibrium"
    for name in names:
        if name in not_continuous_limits:
            continue
        elif name in not_finite_limits:
            assert (np.isfinite(data[name]).T != axis).all(), name
            continue
        else:
            assert np.isfinite(data[name]).all(), name

        if (
            data_index[p][name]["coordinates"] == ""
            or data_index[p][name]["coordinates"] == "z"
        ):
            # can't check radial continuity of scalar or function of toroidal angle
            continue
        # make single variable function of rho
        if data_index[p][name]["coordinates"] == "r":
            # already single variable function of rho
            profile = grid.compress(data[name])
        else:
            # integrate out theta and zeta dependence
            # Norms and integrals are continuous functions, so their composition
            # cannot disrupt existing continuity. Note that the absolute value
            # before the integration ensures that a discontinuous integrand does
            # not become continuous once integrated.
            profile = integrate(np.abs(data[name]))
        fit = kwargs.get(name, {}).get("desired_at_axis", desired_at_axis)
        if fit is None:
            if np.ndim(data_index[p][name]["dim"]):
                # can't polyfit tensor arrays like grad(B)
                fit = profile[1]
            else:
                # fit the data to a polynomial to extrapolate to axis
                poly = np.polynomial.polynomial.polyfit(
                    rho[1:], profile[1:], deg=min(4, num_points // 3)
                )
                # constant term is the same as evaluating polynomial at rho=0
                fit = poly[0]
        np.testing.assert_allclose(
            actual=profile[0],
            desired=fit,
            rtol=kwargs.get(name, {}).get("rtol", rtol),
            atol=kwargs.get(name, {}).get("atol", atol),
            equal_nan=False,
            err_msg=name,
        )


class TestAxisLimits:
    """Tests for compute functions evaluated at limits."""

    @pytest.mark.unit
    def test_axis_limit_api(self):
        """Test that axis limit dependencies are computed only when necessary."""
        name = "psi_r/sqrt(g)"
        deps = {"psi_r", "sqrt(g)"}
        axis_limit_deps = {"psi_rr", "sqrt(g)_r"}
        eq = Equilibrium()
        grid = LinearGrid(L=2, M=1, N=1, axis=False)
        assert not grid.axis.size
        data = eq.compute(name, grid=grid).keys()
        assert name in data and deps < data and axis_limit_deps.isdisjoint(data)
        grid = LinearGrid(L=2, M=1, N=1, axis=True)
        assert grid.axis.size
        data = eq.compute(name, grid=grid)
        assert name in data and deps | axis_limit_deps < data.keys()
        assert np.isfinite(data[name]).all()

    @pytest.mark.unit
    def test_limit_continuity(self):
        """Heuristic to test correctness of all quantities with limits."""
        # The need for a weaker tolerance on these keys may be due to a subpar
        # polynomial regression fit against which the axis limit is compared.
        weaker_tolerance = {
            "iota_r": {"atol": 1e-6},
            "iota_num_rr": {"atol": 5e-5},
            "grad(B)": {"rtol": 1e-4},
        }
        zero_map = dict.fromkeys(zero_limits, {"desired_at_axis": 0})
        kwargs = weaker_tolerance | zero_map
        # fixed iota
        eq = get("W7-X")
        with pytest.warns(UserWarning, match="Reducing radial"):
            eq.change_resolution(4, 4, 4, 8, 8, 8)
        assert_is_continuous(eq, kwargs=kwargs)
        # fixed current
        eq = get("NCSX")
        with pytest.warns(UserWarning, match="Reducing radial"):
            eq.change_resolution(4, 4, 4, 8, 8, 8)
        assert_is_continuous(eq, kwargs=kwargs)

    @pytest.mark.unit
    def test_magnetic_field_is_physical(self):
        """Test direction of magnetic field at axis limit."""

        def test(eq):
            grid = LinearGrid(rho=0, M=5, N=5, NFP=eq.NFP, sym=eq.sym)
            assert grid.axis.size
            data = eq.compute(
                ["b", "n_theta", "n_rho", "e_zeta", "g_zz", "B"], grid=grid
            )
            # For the rotational transform to be finite at the magnetic axis,
            # the magnetic field must satisfy 𝐁 ⋅ 𝐞_ζ × 𝐞ᵨ = 0. This is also
            # required for 𝐁^θ component of the field to be physical.
            np.testing.assert_allclose(dot(data["b"], data["n_theta"]), 0, atol=1e-15)
            # and be orthogonal with 𝐞^ρ because 𝐞^ρ is multivalued at the
            # magnetic axis. 𝐁^ρ = 𝐁 ⋅ 𝐞^ρ must be single-valued for the
            # magnetic field to be physical. (The direction of the vector needs
            # to be unique).
            np.testing.assert_allclose(dot(data["b"], data["n_rho"]), 0, atol=1e-15)
            # and collinear with 𝐞_ζ near ρ=0
            np.testing.assert_allclose(
                # |𝐁_ζ| == ‖𝐁‖ ‖𝐞_ζ‖
                np.abs(dot(data["b"], (data["e_zeta"].T / np.sqrt(data["g_zz"])).T)),
                1,
            )
            # Explicitly check 𝐁 is single-valued at the magnetic axis.
            for B in data["B"].reshape((grid.num_zeta, -1, 3)):
                np.testing.assert_allclose(B[:, 0], B[0, 0])
                np.testing.assert_allclose(B[:, 1], B[0, 1])
                np.testing.assert_allclose(B[:, 2], B[0, 2])

        eq = get("W7-X")
        with pytest.warns(UserWarning, match="Reducing radial"):
            eq.change_resolution(4, 4, 4, 8, 8, 8)
        test(eq)
        eq = get("NCSX")
        with pytest.warns(UserWarning, match="Reducing radial"):
            eq.change_resolution(4, 4, 4, 8, 8, 8)
        test(eq)


def _reverse_mode_unsafe_names():
    names = data_index["desc.equilibrium.equilibrium.Equilibrium"].keys()
    eq = get("ESTELL")

    def isalias(name):
        return isinstance(
            data_index["desc.equilibrium.equilibrium.Equilibrium"][name]["fun"],
            functools.partial,
        )

    def get_source(name):
        return "".join(
            inspect.getsource(
                data_index["desc.equilibrium.equilibrium.Equilibrium"][name]["fun"]
            ).split("def ")[1:]
        )

    names = [
        name
        for name in names
        if not (
            "Boozer" in name
            or "_mn" in name
            or name == "B modes"
            or _skip_this(eq, name)
            or name in not_finite_limits
            or isalias(name)
        )
    ]

    unsafe_names = []  # things that might have nan gradient but shouldn't
    for name in names:
        source = get_source(name)
        if "replace_at_axis" in source:
            unsafe_names.append(name)

    unsafe_names = sorted(unsafe_names)
    return unsafe_names


@pytest.mark.slow
@pytest.mark.regression
@pytest.mark.parametrize("name", _reverse_mode_unsafe_names())
def test_reverse_mode_ad_axis(name):
    """Asserts that the rho=0 axis limits are reverse mode differentiable."""
    eq = get("ESTELL")
    grid = LinearGrid(rho=0.0, M=2, N=2, NFP=eq.NFP, sym=eq.sym)
    with pytest.warns(UserWarning, match="Reducing radial"):
        eq.change_resolution(2, 2, 2, 4, 4, 4)

    obj = ObjectiveFunction(GenericObjective(name, eq, grid=grid), use_jit=False)
    obj.build(verbose=0)
    g = obj.grad(obj.x())
    assert not np.any(np.isnan(g))
    print(np.count_nonzero(g), name)



"""Fast transformable basis."""

from functools import partial

from desc.backend import flatnonzero, jnp, put
from desc.utils import setdefault


@partial(jnp.vectorize, signature="(m),(m)->(m)")
def _in_epigraph_and(is_intersect, df_dy_sign, /):
    """Set and epigraph of function f with the given set of points.

    Used to return only intersects where the straight line path between
    adjacent intersects resides in the epigraph of a continuous map ``f``.

    Parameters
    ----------
    is_intersect : jnp.ndarray
        Boolean array indicating whether index corresponds to an intersect.
    df_dy_sign : jnp.ndarray
        Shape ``is_intersect.shape``.
        Sign of ∂f/∂y (yᵢ) for f(yᵢ) = 0.

    Returns
    -------
    is_intersect : jnp.ndarray
        Boolean array indicating whether element is an intersect
        and satisfies the stated condition.

    Examples
    --------
    See ``desc/integrals/bounce_utils.py::bounce_points``.
    This is used there to ensure the domains of integration are magnetic wells.

    """
    # The pairs ``y1`` and ``y2`` are boundaries of an integral only if ``y1 <= y2``.
    # For the integrals to be over wells, it is required that the first intersect
    # has a non-positive derivative. Now, by continuity,
    # ``df_dy_sign[...,k]<=0`` implies ``df_dy_sign[...,k+1]>=0``,
    # so there can be at most one inversion, and if it exists, the inversion
    # must be at the first pair. To correct the inversion, it suffices to disqualify the
    # first intersect as a right boundary, except under an edge case of a series of
    # inflection points.
    idx = flatnonzero(is_intersect, size=2, fill_value=-1)
    edge_case = (
        (df_dy_sign[idx[0]] == 0)
        & (df_dy_sign[idx[1]] < 0)
        & is_intersect[idx[0]]
        & is_intersect[idx[1]]
        # In theory, we need to keep propagating this edge case, e.g.
        # (df_dy_sign[..., 1] < 0) | (
        #     (df_dy_sign[..., 1] == 0) & (df_dy_sign[..., 2] < 0)...
        # ).
        # At each step, the likelihood that an intersection has already been lost
        # due to floating point errors grows, so the real solution is to pick a less
        # degenerate pitch value - one that does not ride the global extrema of f.
    )
    return put(is_intersect, idx[0], edge_case)


def _add2legend(legend, lines):
    """Add lines to legend if it's not already in it."""
    for line in setdefault(lines, [lines], hasattr(lines, "__iter__")):
        label = line.get_label()
        if label not in legend:
            legend[label] = line


def _plot_intersect(ax, legend, z1, z2, k, k_transparency, klabel):
    """Plot intersects on ``ax``."""
    if k is None:
        return

    k = jnp.atleast_1d(jnp.squeeze(k))
    assert k.ndim == 1
    z1, z2 = jnp.atleast_2d(z1, z2)
    assert z1.ndim == z2.ndim >= 2
    assert k.shape[0] == z1.shape[0] == z2.shape[0]
    for p in k:
        _add2legend(
            legend,
            ax.axhline(p, color="tab:purple", alpha=k_transparency, label=klabel),
        )
    for i in range(k.size):
        _z1, _z2 = z1[i], z2[i]
        if _z1.size == _z2.size:
            mask = (_z1 - _z2) != 0.0
            _z1 = _z1[mask]
            _z2 = _z2[mask]
        _add2legend(
            legend,
            ax.scatter(
                _z1,
                jnp.full_like(_z1, k[i]),
                marker="v",
                color="tab:red",
                label=r"$z_1$",
            ),
        )
        _add2legend(
            legend,
            ax.scatter(
                _z2,
                jnp.full_like(_z2, k[i]),
                marker="^",
                color="tab:green",
                label=r"$z_2$",
            ),
        )



"""Compute functions related to the metric tensor of the coordinate system.

Notes
-----
Some quantities require additional work to compute at the magnetic axis.
A Python lambda function is used to lazily compute the magnetic axis limits
of these quantities. These lambda functions are evaluated only when the
computational grid has a node on the magnetic axis to avoid potentially
expensive computations.
"""

from scipy.constants import mu_0

from desc.backend import jnp

from ..integrals.surface_integral import surface_averages
from ..utils import cross, dot, safediv, safenorm
from .data_index import register_compute_fun


@register_compute_fun(
    name="sqrt(g)",
    label="\\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta", "e_zeta"],
)
def _sqrtg(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)"] = dot(data["e_rho"], cross(data["e_theta"], data["e_zeta"]))
    return data


@register_compute_fun(
    name="sqrt(g)_PEST",
    label="\\sqrt{g}_{PEST}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of (ρ,ϑ,ϕ) coordinate system or"
    " straight field line PEST coordinates. ϕ increases counterclockwise"
    " when viewed from above (cylindrical R,ϕ plane with Z out of page).",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["sqrt(g)", "theta_PEST_t", "phi_z", "theta_PEST_z", "phi_t"],
)
def _sqrtg_pest(params, transforms, profiles, data, **kwargs):
    # Same as dot(data["e_rho|v,p"], cross(data["e_vartheta"], data["e_phi|r,v"])), but
    # more efficient as it avoids computing radial derivatives of the stream functions.
    data["sqrt(g)_PEST"] = data["sqrt(g)"] / (
        data["theta_PEST_t"] * data["phi_z"] - data["theta_PEST_z"] * data["phi_t"]
    )
    return data


@register_compute_fun(
    name="sqrt(g)_Clebsch",
    label="\\sqrt{g}_{\\text{Clebsch}}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of Clebsch field line coordinate system (ρ,α,ζ)"
    " where ζ is the DESC toroidal coordinate.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["sqrt(g)", "alpha_t"],
)
def _sqrtg_clebsch(params, transforms, profiles, data, **kwargs):
    # Same as dot(data["e_rho|a,z"], cross(data["e_alpha"], data["e_zeta|r,a"])), but
    # more efficient as it avoids computing radial derivative of alpha and hence iota.
    data["sqrt(g)_Clebsch"] = data["sqrt(g)"] / data["alpha_t"]
    return data


@register_compute_fun(
    name="|e_theta x e_zeta|",
    label="| \\mathbf{e}_{\\theta} \\times \\mathbf{e}_{\\zeta} |",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant rho surface",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_zeta"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_theta_x_e_zeta(params, transforms, profiles, data, **kwargs):
    data["|e_theta x e_zeta|"] = safenorm(
        cross(data["e_theta"], data["e_zeta"]), axis=-1
    )
    return data


@register_compute_fun(
    name="|e_theta x e_zeta|_r",
    label="\\partial_{\\rho} |\\mathbf{e}_{\\theta} \\times \\mathbf{e}_{\\zeta}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant rho surface"
    + " derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_zeta", "e_theta_r", "e_zeta_r"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_theta_x_e_zeta_r(params, transforms, profiles, data, **kwargs):
    a = cross(data["e_theta"], data["e_zeta"])
    a_r = cross(data["e_theta_r"], data["e_zeta"]) + cross(
        data["e_theta"], data["e_zeta_r"]
    )
    # The limit of a sequence and the norm function can be interchanged
    # because norms are continuous functions. Likewise with dot product.
    # Then lim ‖𝐞^ρ‖ = ‖ lim 𝐞^ρ ‖ ≠ 0
    # lim (𝐞^ρ ⋅ a_r / ‖𝐞^ρ‖) = lim 𝐞^ρ ⋅ lim a_r / lim ‖𝐞^ρ‖
    # The vectors converge to be parallel.
    data["|e_theta x e_zeta|_r"] = transforms["grid"].replace_at_axis(
        safediv(dot(a, a_r), safenorm(a, axis=-1)), lambda: safenorm(a_r, axis=-1)
    )
    return data


@register_compute_fun(
    name="|e_theta x e_zeta|_rr",
    label="\\partial_{\\rho\\rho}|\\mathbf{e}_{\\theta}\\times\\mathbf{e}_{\\zeta}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant rho surface"
    + " second derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_zeta", "e_theta_r", "e_zeta_r", "e_theta_rr", "e_zeta_rr"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_theta_x_e_zeta_rr(params, transforms, profiles, data, **kwargs):
    a = cross(data["e_theta"], data["e_zeta"])
    a_r = cross(data["e_theta_r"], data["e_zeta"]) + cross(
        data["e_theta"], data["e_zeta_r"]
    )
    a_rr = (
        cross(data["e_theta_rr"], data["e_zeta"])
        + 2 * cross(data["e_theta_r"], data["e_zeta_r"])
        + cross(data["e_theta"], data["e_zeta_rr"])
    )
    norm_a = safenorm(a, axis=-1)
    norm_a_r = safenorm(a_r, axis=-1)
    # The limit eventually reduces to a form where the technique used to compute
    # lim |e_theta x e_zeta|_r can be applied.
    data["|e_theta x e_zeta|_rr"] = transforms["grid"].replace_at_axis(
        safediv(norm_a_r**2 + dot(a, a_rr) - safediv(dot(a, a_r), norm_a) ** 2, norm_a),
        lambda: safediv(dot(a_r, a_rr), norm_a_r),
    )
    return data


@register_compute_fun(
    name="|e_theta x e_zeta|_z",
    label="\\partial_{\\zeta}|\\mathbf{e}_{\\theta} \\times \\mathbf{e}_{\\zeta}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant rho surface,"
    "derivative wrt toroidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_theta_z", "e_zeta", "e_zeta_z", "|e_theta x e_zeta|"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_theta_x_e_zeta_z(params, transforms, profiles, data, **kwargs):
    data["|e_theta x e_zeta|_z"] = dot(
        (
            cross(data["e_theta_z"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_z"])
        ),
        cross(data["e_theta"], data["e_zeta"]),
    ) / (data["|e_theta x e_zeta|"])
    return data


@register_compute_fun(
    name="|e_zeta x e_rho|",
    label="|\\mathbf{e}_{\\zeta} \\times \\mathbf{e}_{\\rho}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant theta surface",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta*sqrt(g)"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_zeta_x_e_rho(params, transforms, profiles, data, **kwargs):
    data["|e_zeta x e_rho|"] = safenorm(data["e^theta*sqrt(g)"], axis=-1)
    return data


@register_compute_fun(
    name="|e_rho x e_theta|",
    label="|\\mathbf{e}_{\\rho} \\times \\mathbf{e}_{\\theta}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant zeta surface",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_rho_x_e_theta(params, transforms, profiles, data, **kwargs):
    data["|e_rho x e_theta|"] = safenorm(cross(data["e_rho"], data["e_theta"]), axis=-1)
    return data


@register_compute_fun(
    name="|e_rho x e_alpha|",
    label="|\\mathbf{e}_{\\rho} \\times \\mathbf{e}_{\\alpha}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant zeta surface in Clebsch "
    "field line coordinates (ρ,α,ζ) where ζ is the DESC toroidal coordinate.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["|e_rho x e_theta|", "alpha_t"],
)
def _e_rho_x_e_alpha(params, transforms, profiles, data, **kwargs):
    # Same as safenorm(cross(data["e_rho|a,z"], data["e_alpha"]), axis=-1), but more
    # efficient as it avoids computing radial derivative of iota and stream functions.
    data["|e_rho x e_alpha|"] = data["|e_rho x e_theta|"] / jnp.abs(data["alpha_t"])
    return data


@register_compute_fun(
    name="|e_rho x e_theta|_r",
    label="\\partial_{\\rho} |\\mathbf{e}_{\\rho} \\times \\mathbf{e}_{\\theta}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant zeta surface"
    " derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta", "e_rho_r", "e_theta_r"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_rho_x_e_theta_r(params, transforms, profiles, data, **kwargs):
    a = cross(data["e_rho"], data["e_theta"])
    a_r = cross(data["e_rho_r"], data["e_theta"]) + cross(
        data["e_rho"], data["e_theta_r"]
    )
    # The limit of a sequence and the norm function can be interchanged
    # because norms are continuous functions. Likewise with dot product.
    # Then lim ‖𝐞^ζ‖ = ‖ lim 𝐞^ζ ‖ ≠ 0
    # lim (𝐞^ζ ⋅ a_r / ‖𝐞^ζ‖) = lim 𝐞^ζ ⋅ lim a_r / lim ‖𝐞^ζ‖
    # The vectors converge to be parallel.
    data["|e_rho x e_theta|_r"] = transforms["grid"].replace_at_axis(
        safediv(dot(a, a_r), safenorm(a, axis=-1)), lambda: safenorm(a_r, axis=-1)
    )
    return data


@register_compute_fun(
    name="|e_rho x e_theta|_rr",
    label="\\partial_{\\rho \\rho} |\\mathbf{e}_{\\rho} \\times \\mathbf{e}_{\\theta}|",
    units="m^{2}",
    units_long="square meters",
    description="2D Jacobian determinant for constant zeta surface"
    + " second derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta", "e_rho_r", "e_theta_r", "e_rho_rr", "e_theta_rr"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _e_rho_x_e_theta_rr(params, transforms, profiles, data, **kwargs):
    a = cross(data["e_rho"], data["e_theta"])
    a_r = cross(data["e_rho_r"], data["e_theta"]) + cross(
        data["e_rho"], data["e_theta_r"]
    )
    a_rr = (
        cross(data["e_rho_rr"], data["e_theta"])
        + 2 * cross(data["e_rho_r"], data["e_theta_r"])
        + cross(data["e_rho"], data["e_theta_rr"])
    )
    norm_a = safenorm(a, axis=-1)
    norm_a_r = safenorm(a_r, axis=-1)
    # The limit eventually reduces to a form where the technique used to compute
    # lim |e_rho x e_theta|_r can be applied.
    data["|e_rho x e_theta|_rr"] = transforms["grid"].replace_at_axis(
        safediv(norm_a_r**2 + dot(a, a_rr) - safediv(dot(a, a_r), norm_a) ** 2, norm_a),
        lambda: safediv(dot(a_r, a_rr), norm_a_r),
    )
    return data


@register_compute_fun(
    name="sqrt(g)_r",
    label="\\partial_{\\rho} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, derivative wrt "
    + "radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta", "e_zeta", "e_rho_r", "e_theta_r", "e_zeta_r"],
)
def _sqrtg_r(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_r"] = (
        dot(data["e_rho_r"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_r"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_r"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_t",
    label="\\partial_{\\theta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, derivative wrt "
    + "poloidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta", "e_zeta", "e_rho_t", "e_theta_t", "e_zeta_t"],
)
def _sqrtg_t(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_t"] = (
        dot(data["e_rho_t"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_t"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_t"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_z",
    label="\\partial_{\\zeta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, derivative wrt "
    + "toroidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta", "e_zeta", "e_rho_z", "e_theta_z", "e_zeta_z"],
)
def _sqrtg_z(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_z"] = (
        dot(data["e_rho_z"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_z"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_z"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rr",
    label="\\partial_{\\rho\\rho} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, second derivative wrt "
    + "radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_rr",
        "e_theta_rr",
        "e_zeta_rr",
    ],
)
def _sqrtg_rr(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rr"] = (
        dot(data["e_rho_rr"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_rr"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_rr"]))
        + 2 * dot(data["e_rho_r"], cross(data["e_theta_r"], data["e_zeta"]))
        + 2 * dot(data["e_rho_r"], cross(data["e_theta"], data["e_zeta_r"]))
        + 2 * dot(data["e_rho"], cross(data["e_theta_r"], data["e_zeta_r"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rrr",
    label="\\partial_{\\rho\\rho\\rho} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, third derivative wrt "
    + "radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_rr",
        "e_theta_rr",
        "e_zeta_rr",
        "e_rho_rrr",
        "e_theta_rrr",
        "e_zeta_rrr",
    ],
)
def _sqrtg_rrr(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rrr"] = (
        dot(data["e_rho_rrr"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_rrr"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_rrr"]))
        + 3 * dot(data["e_rho_rr"], cross(data["e_theta_r"], data["e_zeta"]))
        + 3 * dot(data["e_rho_rr"], cross(data["e_theta"], data["e_zeta_r"]))
        + 3 * dot(data["e_rho_r"], cross(data["e_theta_rr"], data["e_zeta"]))
        + 3 * dot(data["e_rho"], cross(data["e_theta_rr"], data["e_zeta_r"]))
        + 3 * dot(data["e_rho_r"], cross(data["e_theta"], data["e_zeta_rr"]))
        + 3 * dot(data["e_rho"], cross(data["e_theta_r"], data["e_zeta_rr"]))
        + 6 * dot(data["e_rho_r"], cross(data["e_theta_r"], data["e_zeta_r"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rrt",
    label="\\partial_{\\rho\\rho\\theta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, third derivative wrt "
    + "radial coordinate twice and poloidal angle once",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_t",
        "e_theta_t",
        "e_zeta_t",
        "e_rho_rt",
        "e_theta_rt",
        "e_zeta_rt",
        "e_rho_rr",
        "e_theta_rr",
        "e_zeta_rr",
        "e_rho_rrt",
        "e_theta_rrt",
        "e_zeta_rrt",
    ],
)
def _sqrtg_rrt(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rrt"] = (
        dot(data["e_rho_rrt"], cross(data["e_theta"], data["e_zeta"]))
        + dot(
            data["e_rho_rr"],
            cross(data["e_theta_t"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_t"]),
        )
        + 2
        * dot(
            data["e_rho_rt"],
            cross(data["e_theta_r"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_r"]),
        )
        + 2
        * dot(
            data["e_rho_r"],
            cross(data["e_theta_rt"], data["e_zeta"])
            + cross(data["e_theta_r"], data["e_zeta_t"])
            + cross(data["e_theta_t"], data["e_zeta_r"])
            + cross(data["e_theta"], data["e_zeta_rt"]),
        )
        + dot(
            data["e_rho_t"],
            cross(data["e_theta_rr"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_rr"]),
        )
        + dot(
            data["e_rho"],
            cross(data["e_theta_rrt"], data["e_zeta"])
            + 2 * cross(data["e_theta_rt"], data["e_zeta_r"])
            + cross(data["e_theta_rr"], data["e_zeta_t"])
            + 2 * cross(data["e_theta_r"], data["e_zeta_rt"])
            + cross(data["e_theta_t"], data["e_zeta_rr"])
            + cross(data["e_theta"], data["e_zeta_rrt"]),
        )
    )
    return data


@register_compute_fun(
    name="sqrt(g)_tt",
    label="\\partial_{\\theta\\theta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, second derivative wrt "
    + "poloidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_t",
        "e_theta_t",
        "e_zeta_t",
        "e_rho_tt",
        "e_theta_tt",
        "e_zeta_tt",
    ],
)
def _sqrtg_tt(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_tt"] = (
        dot(data["e_rho_tt"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_tt"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_tt"]))
        + 2 * dot(data["e_rho_t"], cross(data["e_theta_t"], data["e_zeta"]))
        + 2 * dot(data["e_rho_t"], cross(data["e_theta"], data["e_zeta_t"]))
        + 2 * dot(data["e_rho"], cross(data["e_theta_t"], data["e_zeta_t"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rtt",
    label="\\partial_{\\rho\\theta\\theta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, third derivative wrt"
    + " radial coordinate once and poloidal angle twice.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_t",
        "e_theta_t",
        "e_zeta_t",
        "e_rho_rt",
        "e_theta_rt",
        "e_zeta_rt",
        "e_rho_tt",
        "e_theta_tt",
        "e_zeta_tt",
        "e_rho_rtt",
        "e_theta_rtt",
        "e_zeta_rtt",
    ],
)
def _sqrtg_rtt(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rtt"] = (
        dot(data["e_rho_rtt"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta_tt"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta"], data["e_zeta_tt"]))
        + 2 * dot(data["e_rho_rt"], cross(data["e_theta_t"], data["e_zeta"]))
        + 2 * dot(data["e_rho_rt"], cross(data["e_theta"], data["e_zeta_t"]))
        + 2 * dot(data["e_rho_r"], cross(data["e_theta_t"], data["e_zeta_t"]))
        + dot(data["e_rho_tt"], cross(data["e_theta_r"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_rtt"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_r"], data["e_zeta_tt"]))
        + 2 * dot(data["e_rho_t"], cross(data["e_theta_rt"], data["e_zeta"]))
        + 2 * dot(data["e_rho"], cross(data["e_theta_rt"], data["e_zeta_t"]))
        + dot(data["e_rho_tt"], cross(data["e_theta"], data["e_zeta_r"]))
        + dot(data["e_rho"], cross(data["e_theta_tt"], data["e_zeta_r"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_rtt"]))
        + 2 * dot(data["e_rho_t"], cross(data["e_theta_t"], data["e_zeta_r"]))
        + 2 * dot(data["e_rho_t"], cross(data["e_theta"], data["e_zeta_rt"]))
        + 2 * dot(data["e_rho"], cross(data["e_theta_t"], data["e_zeta_rt"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_zz",
    label="\\partial_{\\zeta\\zeta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, second derivative wrt "
    + "toroidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_z",
        "e_theta_z",
        "e_zeta_z",
        "e_rho_zz",
        "e_theta_zz",
        "e_zeta_zz",
    ],
)
def _sqrtg_zz(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_zz"] = (
        dot(data["e_rho_zz"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_zz"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_zz"]))
        + 2 * dot(data["e_rho_z"], cross(data["e_theta_z"], data["e_zeta"]))
        + 2 * dot(data["e_rho_z"], cross(data["e_theta"], data["e_zeta_z"]))
        + 2 * dot(data["e_rho"], cross(data["e_theta_z"], data["e_zeta_z"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rzz",
    label="\\partial_{\\rho\\zeta\\zeta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, third derivative wrt "
    + "radial coordinate once and toroidal angle twice",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_z",
        "e_theta_z",
        "e_zeta_z",
        "e_rho_zz",
        "e_theta_zz",
        "e_zeta_zz",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_rz",
        "e_theta_rz",
        "e_zeta_rz",
        "e_rho_rzz",
        "e_theta_rzz",
        "e_zeta_rzz",
    ],
)
def _sqrtg_rzz(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rzz"] = (
        dot(data["e_rho_rzz"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta_zz"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta"], data["e_zeta_zz"]))
        + 2 * dot(data["e_rho_rz"], cross(data["e_theta_z"], data["e_zeta"]))
        + 2 * dot(data["e_rho_rz"], cross(data["e_theta"], data["e_zeta_z"]))
        + 2 * dot(data["e_rho_r"], cross(data["e_theta_z"], data["e_zeta_z"]))
        + dot(data["e_rho_zz"], cross(data["e_theta_r"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_rzz"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_r"], data["e_zeta_zz"]))
        + 2 * dot(data["e_rho_z"], cross(data["e_theta_rz"], data["e_zeta"]))
        + 2 * dot(data["e_rho_z"], cross(data["e_theta_r"], data["e_zeta_z"]))
        + 2 * dot(data["e_rho"], cross(data["e_theta_rz"], data["e_zeta_z"]))
        + dot(data["e_rho_zz"], cross(data["e_theta"], data["e_zeta_r"]))
        + dot(data["e_rho"], cross(data["e_theta_zz"], data["e_zeta_r"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_rzz"]))
        + 2 * dot(data["e_rho_z"], cross(data["e_theta"], data["e_zeta_rz"]))
        + 2 * dot(data["e_rho"], cross(data["e_theta_z"], data["e_zeta_rz"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rt",
    label="\\partial_{\\rho\\theta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, second derivative wrt "
    + "radial coordinate and poloidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_t",
        "e_theta_t",
        "e_zeta_t",
        "e_rho_rt",
        "e_theta_rt",
        "e_zeta_rt",
    ],
)
def _sqrtg_rt(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rt"] = (
        dot(data["e_rho_rt"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta_t"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta"], data["e_zeta_t"]))
        + dot(data["e_rho"], cross(data["e_theta_rt"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_r"], data["e_zeta_t"]))
        + dot(data["e_rho_t"], cross(data["e_theta"], data["e_zeta_r"]))
        + dot(data["e_rho"], cross(data["e_theta_t"], data["e_zeta_r"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_rt"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_tz",
    label="\\partial_{\\theta\\zeta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, second derivative wrt "
    + "poloidal and toroidal angles",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_t",
        "e_theta_t",
        "e_zeta_t",
        "e_rho_z",
        "e_theta_z",
        "e_zeta_z",
        "e_rho_tz",
        "e_theta_tz",
        "e_zeta_tz",
    ],
)
def _sqrtg_tz(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_tz"] = (
        dot(data["e_rho_tz"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho_z"], cross(data["e_theta_t"], data["e_zeta"]))
        + dot(data["e_rho_z"], cross(data["e_theta"], data["e_zeta_t"]))
        + dot(data["e_rho_t"], cross(data["e_theta_z"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_tz"], data["e_zeta"]))
        + dot(data["e_rho_t"], cross(data["e_theta"], data["e_zeta_z"]))
        + dot(data["e_rho"], cross(data["e_theta_t"], data["e_zeta_z"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_tz"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rtz",
    label="\\partial_{\\rho\\theta\\zeta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, third derivative wrt "
    + "radial, poloidal, and toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_t",
        "e_theta_t",
        "e_zeta_t",
        "e_rho_rt",
        "e_theta_rt",
        "e_zeta_rt",
        "e_rho_z",
        "e_theta_z",
        "e_zeta_z",
        "e_rho_rz",
        "e_theta_rz",
        "e_zeta_rz",
        "e_rho_tz",
        "e_theta_tz",
        "e_zeta_tz",
        "e_rho_rtz",
        "e_theta_rtz",
        "e_zeta_rtz",
    ],
)
def _sqrtg_rtz(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rtz"] = (
        dot(data["e_rho_rtz"], cross(data["e_theta"], data["e_zeta"]))
        + dot(
            data["e_rho_rz"],
            cross(data["e_theta_t"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_t"]),
        )
        + dot(
            data["e_rho_rt"],
            cross(data["e_theta_z"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_z"]),
        )
        + dot(
            data["e_rho_r"],
            cross(data["e_theta_tz"], data["e_zeta"])
            + cross(data["e_theta_t"], data["e_zeta_z"])
            + cross(data["e_theta"], data["e_zeta_tz"]),
        )
        + dot(
            data["e_rho_tz"],
            cross(data["e_theta_r"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_r"]),
        )
        + dot(
            data["e_rho_z"],
            cross(data["e_theta_rt"], data["e_zeta"])
            + cross(data["e_theta_r"], data["e_zeta_t"])
            + cross(data["e_theta"], data["e_zeta_rt"]),
        )
        + dot(
            data["e_rho_t"],
            cross(data["e_theta_rz"], data["e_zeta"])
            + cross(data["e_theta_z"], data["e_zeta_r"])
            + cross(data["e_theta"], data["e_zeta_rz"]),
        )
        + dot(
            data["e_rho"],
            cross(data["e_theta_rtz"], data["e_zeta"])
            + cross(data["e_theta_tz"], data["e_zeta_r"])
            + cross(data["e_theta_rz"], data["e_zeta_t"])
            + cross(data["e_theta_z"], data["e_zeta_rt"])
            + cross(data["e_theta_rt"], data["e_zeta_z"])
            + cross(data["e_theta_t"], data["e_zeta_rz"])
            + cross(data["e_theta_r"], data["e_zeta_tz"])
            + cross(data["e_theta"], data["e_zeta_rtz"]),
        )
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rz",
    label="\\partial_{\\rho\\zeta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, second derivative wrt "
    + "radial coordinate and toroidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_z",
        "e_theta_z",
        "e_zeta_z",
        "e_rho_rz",
        "e_theta_rz",
        "e_zeta_rz",
    ],
)
def _sqrtg_rz(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rz"] = (
        dot(data["e_rho_rz"], cross(data["e_theta"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta_z"], data["e_zeta"]))
        + dot(data["e_rho_r"], cross(data["e_theta"], data["e_zeta_z"]))
        + dot(data["e_rho_z"], cross(data["e_theta_r"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_rz"], data["e_zeta"]))
        + dot(data["e_rho"], cross(data["e_theta_r"], data["e_zeta_z"]))
        + dot(data["e_rho"], cross(data["e_theta_z"], data["e_zeta_r"]))
        + dot(data["e_rho"], cross(data["e_theta"], data["e_zeta_rz"]))
    )
    return data


@register_compute_fun(
    name="sqrt(g)_rrz",
    label="\\partial_{\\rho\\rho\\zeta} \\sqrt{g}",
    units="m^{3}",
    units_long="cubic meters",
    description="Jacobian determinant of flux coordinate system, third derivative wrt "
    + "radial coordinate twice and toroidal angle once",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
        "e_rho_z",
        "e_theta_z",
        "e_zeta_z",
        "e_rho_rr",
        "e_rho_rz",
        "e_theta_rr",
        "e_theta_rz",
        "e_zeta_rz",
        "e_zeta_rr",
        "e_rho_rrz",
        "e_theta_rrz",
        "e_zeta_rrz",
    ],
)
def _sqrtg_rrz(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_rrz"] = (
        dot(data["e_rho_rrz"], cross(data["e_theta"], data["e_zeta"]))
        + dot(
            data["e_rho_rr"],
            cross(data["e_theta_z"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_z"]),
        )
        + 2
        * dot(
            data["e_rho_rz"],
            cross(data["e_theta_r"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_r"]),
        )
        + 2
        * dot(
            data["e_rho_r"],
            cross(data["e_theta_rz"], data["e_zeta"])
            + cross(data["e_theta_r"], data["e_zeta_z"])
            + cross(data["e_theta_z"], data["e_zeta_r"])
            + cross(data["e_theta"], data["e_zeta_rz"]),
        )
        + dot(
            data["e_rho_z"],
            cross(data["e_theta_rr"], data["e_zeta"])
            + cross(data["e_theta"], data["e_zeta_rr"]),
        )
        + dot(
            data["e_rho"],
            cross(data["e_theta_rrz"], data["e_zeta"])
            + cross(data["e_theta_rr"], data["e_zeta_z"])
            + 2 * cross(data["e_theta_r"], data["e_zeta_rz"])
            + 2 * cross(data["e_theta_rz"], data["e_zeta_r"])
            + cross(data["e_theta_z"], data["e_zeta_rr"])
            + cross(data["e_theta"], data["e_zeta_rrz"]),
        )
    )
    return data


@register_compute_fun(
    name="g_rr",
    label="g_{\\rho\\rho}",
    units="m^{2}",
    units_long="square meters",
    description="Radial/Radial element of covariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_rr(params, transforms, profiles, data, **kwargs):
    data["g_rr"] = dot(data["e_rho"], data["e_rho"])
    return data


@register_compute_fun(
    name="g_tt",
    label="g_{\\theta\\theta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Poloidal element of covariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_tt(params, transforms, profiles, data, **kwargs):
    data["g_tt"] = dot(data["e_theta"], data["e_theta"])
    return data


@register_compute_fun(
    name="g_zz",
    label="g_{\\zeta\\zeta}",
    units="m^{2}",
    units_long="square meters",
    description="Toroidal/Toroidal element of covariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_zeta"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_zz(params, transforms, profiles, data, **kwargs):
    data["g_zz"] = dot(data["e_zeta"], data["e_zeta"])
    return data


@register_compute_fun(
    name="g_rt",
    label="g_{\\rho\\theta}",
    units="m^{2}",
    units_long="square meters",
    description="Radial/Poloidal element of covariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_theta"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_rt(params, transforms, profiles, data, **kwargs):
    data["g_rt"] = dot(data["e_rho"], data["e_theta"])
    return data


@register_compute_fun(
    name="g_rz",
    label="g_{\\rho\\zeta}",
    units="m^{2}",
    units_long="square meters",
    description="Radial/Toroidal element of covariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_rho", "e_zeta"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_rz(params, transforms, profiles, data, **kwargs):
    data["g_rz"] = dot(data["e_rho"], data["e_zeta"])
    return data


@register_compute_fun(
    name="g_tz",
    label="g_{\\theta\\zeta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Toroidal element of covariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_zeta"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_tz(params, transforms, profiles, data, **kwargs):
    data["g_tz"] = dot(data["e_theta"], data["e_zeta"])
    return data


@register_compute_fun(
    name="g_tt_r",
    label="\\partial_{\\rho} g_{\\theta\\theta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Poloidal element of covariant metric tensor, derivative "
    + "wrt rho",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_theta_r"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_tt_r(params, transforms, profiles, data, **kwargs):
    data["g_tt_r"] = 2 * dot(data["e_theta"], data["e_theta_r"])
    return data


@register_compute_fun(
    name="g_tz_r",
    label="\\partial_{\\rho} g_{\\theta\\zeta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Toroidal element of covariant metric tensor, derivative "
    + "wrt rho",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_zeta", "e_theta_r", "e_zeta_r"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _g_sub_tz_r(params, transforms, profiles, data, **kwargs):
    data["g_tz_r"] = dot(data["e_theta_r"], data["e_zeta"]) + dot(
        data["e_theta"], data["e_zeta_r"]
    )
    return data


@register_compute_fun(
    name="g_tt_rr",
    label="\\partial_{\\rho\\rho} g_{\\theta\\theta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Poloidal element of covariant metric tensor, second "
    + "derivative wrt rho",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_theta_r", "e_theta_rr"],
)
def _g_sub_tt_rr(params, transforms, profiles, data, **kwargs):
    data["g_tt_rr"] = 2 * (
        dot(data["e_theta_r"], data["e_theta_r"])
        + dot(data["e_theta"], data["e_theta_rr"])
    )
    return data


@register_compute_fun(
    name="g_tt_rrr",
    label="\\partial_{\\rho\\rho\\rho} g_{\\theta\\theta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Poloidal element of covariant metric tensor, third "
    + "derivative wrt rho",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_theta_r", "e_theta_rr", "e_theta_rrr"],
)
def _g_sub_tt_rrr(params, transforms, profiles, data, **kwargs):
    data["g_tt_rrr"] = 6 * dot(data["e_theta_rr"], data["e_theta_r"]) + 2 * dot(
        data["e_theta"], data["e_theta_rrr"]
    )
    return data


@register_compute_fun(
    name="g_tz_rr",
    label="\\partial_{\\rho\\rho} g_{\\theta\\zeta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Toroidal element of covariant metric tensor, second "
    + "derivative wrt rho",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e_theta", "e_zeta", "e_theta_r", "e_zeta_r", "e_theta_rr", "e_zeta_rr"],
)
def _g_sub_tz_rr(params, transforms, profiles, data, **kwargs):
    data["g_tz_rr"] = (
        dot(data["e_theta_rr"], data["e_zeta"])
        + 2 * dot(data["e_theta_r"], data["e_zeta_r"])
        + dot(data["e_theta"], data["e_zeta_rr"])
    )
    return data


@register_compute_fun(
    name="g_tz_rrr",
    label="\\partial_{\\rho\\rho\\rho} g_{\\theta\\zeta}",
    units="m^{2}",
    units_long="square meters",
    description="Poloidal/Toroidal element of covariant metric tensor, third "
    + "derivative wrt rho",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "e_theta",
        "e_zeta",
        "e_theta_r",
        "e_zeta_r",
        "e_theta_rr",
        "e_zeta_rr",
        "e_theta_rrr",
        "e_zeta_rrr",
    ],
)
def _g_sub_tz_rrr(params, transforms, profiles, data, **kwargs):
    data["g_tz_rrr"] = (
        dot(data["e_theta_rrr"], data["e_zeta"])
        + 3 * dot(data["e_theta_rr"], data["e_zeta_r"])
        + 3 * dot(data["e_theta_r"], data["e_zeta_rr"])
        + dot(data["e_theta"], data["e_zeta_rrr"])
    )
    return data


@register_compute_fun(
    name="g^rr",
    label="g^{\\rho\\rho}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Radial/Radial element of contravariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho"],
)
def _g_sup_rr(params, transforms, profiles, data, **kwargs):
    data["g^rr"] = dot(data["e^rho"], data["e^rho"])
    return data


@register_compute_fun(
    name="g^tt",
    label="g^{\\theta\\theta}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Poloidal/Poloidal element of contravariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta"],
)
def _g_sup_tt(params, transforms, profiles, data, **kwargs):
    data["g^tt"] = dot(data["e^theta"], data["e^theta"])
    return data


@register_compute_fun(
    name="g^zz",
    label="g^{\\zeta\\zeta}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Toroidal/Toroidal element of contravariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^zeta"],
)
def _g_sup_zz(params, transforms, profiles, data, **kwargs):
    data["g^zz"] = dot(data["e^zeta"], data["e^zeta"])
    return data


@register_compute_fun(
    name="g^rt",
    label="g^{\\rho\\theta}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Radial/Poloidal (ρ, θ) element of contravariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^theta"],
)
def _g_sup_rt(params, transforms, profiles, data, **kwargs):
    data["g^rt"] = dot(data["e^rho"], data["e^theta"])
    return data


@register_compute_fun(
    name="g^rz",
    label="g^{\\rho\\zeta}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Radial/Toroidal element of contravariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^zeta"],
)
def _g_sup_rz(params, transforms, profiles, data, **kwargs):
    data["g^rz"] = dot(data["e^rho"], data["e^zeta"])
    return data


@register_compute_fun(
    name="g^tz",
    label="g^{\\theta\\zeta}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Poloidal/Toroidal element of contravariant metric tensor",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta", "e^zeta"],
    aliases=["g^zt"],
)
def _g_sup_tz(params, transforms, profiles, data, **kwargs):
    data["g^tz"] = dot(data["e^theta"], data["e^zeta"])
    return data


@register_compute_fun(
    name="g^rr_r",
    label="\\partial_{\\rho} g^{\\rho \\rho}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Radial element of contravariant metric tensor, "
    + "first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^rho_r"],
)
def _g_sup_rr_r(params, transforms, profiles, data, **kwargs):
    data["g^rr_r"] = 2 * dot(data["e^rho_r"], data["e^rho"])
    return data


@register_compute_fun(
    name="g^rr_t",
    label="\\partial_{\\theta} g^{\\rho \\rho}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Radial element of contravariant metric tensor, "
    + "first poloidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^rho_t"],
)
def _g_sup_rr_t(params, transforms, profiles, data, **kwargs):
    data["g^rr_t"] = 2 * dot(data["e^rho_t"], data["e^rho"])
    return data


@register_compute_fun(
    name="g^rr_z",
    label="\\partial_{\\zeta} g^{\\rho \\rho}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Radial element of contravariant metric tensor, "
    + "first toroidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^rho_z"],
)
def _g_sup_rr_z(params, transforms, profiles, data, **kwargs):
    data["g^rr_z"] = 2 * dot(data["e^rho_z"], data["e^rho"])
    return data


@register_compute_fun(
    name="g^rt_r",
    label="\\partial_{\\rho} g^{\\rho \\theta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Poloidal element of contravariant metric tensor, "
    + "first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^theta", "e^rho_r", "e^theta_r"],
)
def _g_sup_rt_r(params, transforms, profiles, data, **kwargs):
    data["g^rt_r"] = dot(data["e^rho_r"], data["e^theta"]) + dot(
        data["e^rho"], data["e^theta_r"]
    )
    return data


@register_compute_fun(
    name="g^rz_r",
    label="\\partial_{\\rho} g^{\\rho \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Toroidal element of contravariant metric tensor, "
    + "first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^zeta", "e^rho_r", "e^zeta_r"],
)
def _g_sup_rz_r(params, transforms, profiles, data, **kwargs):
    data["g^rz_r"] = dot(data["e^rho_r"], data["e^zeta"]) + dot(
        data["e^rho"], data["e^zeta_r"]
    )
    return data


@register_compute_fun(
    name="g^tt_r",
    label="\\partial_{\\rho} g^{\\theta \\theta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Poloidal/Poloidal element of contravariant metric tensor, "
    + "first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta", "e^theta_r"],
)
def _g_sup_tt_r(params, transforms, profiles, data, **kwargs):
    data["g^tt_r"] = 2 * dot(data["e^theta_r"], data["e^theta"])
    return data


@register_compute_fun(
    name="g^tz_r",
    label="\\partial_{\\rho} g^{\\theta \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Poloidal/Toroidal element of contravariant metric tensor, "
    + "first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta", "e^zeta", "e^theta_r", "e^zeta_r"],
)
def _g_sup_tz_r(params, transforms, profiles, data, **kwargs):
    data["g^tz_r"] = dot(data["e^theta_r"], data["e^zeta"]) + dot(
        data["e^theta"], data["e^zeta_r"]
    )
    return data


@register_compute_fun(
    name="g^zz_r",
    label="\\partial_{\\rho} g^{\\zeta \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Toroidal/Toroidal element of contravariant metric tensor, "
    + "first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^zeta", "e^zeta_r"],
)
def _g_sup_zz_r(params, transforms, profiles, data, **kwargs):
    data["g^zz_r"] = 2 * dot(data["e^zeta_r"], data["e^zeta"])
    return data


@register_compute_fun(
    name="g^rt_t",
    label="\\partial_{\\theta} g^{\\rho \\theta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Poloidal element of contravariant metric tensor, "
    + "first poloidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^theta", "e^rho_t", "e^theta_t"],
)
def _g_sup_rt_t(params, transforms, profiles, data, **kwargs):
    data["g^rt_t"] = dot(data["e^rho_t"], data["e^theta"]) + dot(
        data["e^rho"], data["e^theta_t"]
    )
    return data


@register_compute_fun(
    name="g^rz_t",
    label="\\partial_{\\theta} g^{\\rho \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Toroidal element of contravariant metric tensor, "
    + "first poloidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^zeta", "e^rho_t", "e^zeta_t"],
)
def _g_sup_rz_t(params, transforms, profiles, data, **kwargs):
    data["g^rz_t"] = dot(data["e^rho_t"], data["e^zeta"]) + dot(
        data["e^rho"], data["e^zeta_t"]
    )
    return data


@register_compute_fun(
    name="g^tt_t",
    label="\\partial_{\\theta} g^{\\theta \\theta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Poloidal/Poloidal element of contravariant metric tensor, "
    + "first poloidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta", "e^theta_t"],
)
def _g_sup_tt_t(params, transforms, profiles, data, **kwargs):
    data["g^tt_t"] = 2 * dot(data["e^theta_t"], data["e^theta"])
    return data


@register_compute_fun(
    name="g^tz_t",
    label="\\partial_{\\theta} g^{\\theta \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Poloidal/Toroidal element of contravariant metric tensor, "
    + "first poloidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta", "e^zeta", "e^theta_t", "e^zeta_t"],
)
def _g_sup_tz_t(params, transforms, profiles, data, **kwargs):
    data["g^tz_t"] = dot(data["e^theta_t"], data["e^zeta"]) + dot(
        data["e^theta"], data["e^zeta_t"]
    )
    return data


@register_compute_fun(
    name="g^zz_t",
    label="\\partial_{\\theta} g^{\\zeta \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Toroidal/Toroidal element of contravariant metric tensor, "
    + "first poloidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^zeta", "e^zeta_t"],
)
def _g_sup_zz_t(params, transforms, profiles, data, **kwargs):
    data["g^zz_t"] = 2 * dot(data["e^zeta_t"], data["e^zeta"])
    return data


@register_compute_fun(
    name="g^rt_z",
    label="\\partial_{\\zeta} g^{\\rho \\theta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Poloidal element of contravariant metric tensor, "
    + "first toroidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^theta", "e^rho_z", "e^theta_z"],
)
def _g_sup_rt_z(params, transforms, profiles, data, **kwargs):
    data["g^rt_z"] = dot(data["e^rho_z"], data["e^theta"]) + dot(
        data["e^rho"], data["e^theta_z"]
    )
    return data


@register_compute_fun(
    name="g^rz_z",
    label="\\partial_{\\zeta} g^{\\rho \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Radial/Toroidal element of contravariant metric tensor, "
    + "first toroidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^rho", "e^zeta", "e^rho_z", "e^zeta_z"],
)
def _g_sup_rz_z(params, transforms, profiles, data, **kwargs):
    data["g^rz_z"] = dot(data["e^rho_z"], data["e^zeta"]) + dot(
        data["e^rho"], data["e^zeta_z"]
    )
    return data


@register_compute_fun(
    name="g^tt_z",
    label="\\partial_{\\zeta} g^{\\theta \\theta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Poloidal/Poloidal element of contravariant metric tensor, "
    + "first toroidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta", "e^theta_z"],
)
def _g_sup_tt_z(params, transforms, profiles, data, **kwargs):
    data["g^tt_z"] = 2 * dot(data["e^theta_z"], data["e^theta"])
    return data


@register_compute_fun(
    name="g^tz_z",
    label="\\partial_{\\zeta} g^{\\theta \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Poloidal/Toroidal element of contravariant metric tensor, "
    + "first toroidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^theta", "e^zeta", "e^theta_z", "e^zeta_z"],
)
def _g_sup_tz_z(params, transforms, profiles, data, **kwargs):
    data["g^tz_z"] = dot(data["e^theta_z"], data["e^zeta"]) + dot(
        data["e^theta"], data["e^zeta_z"]
    )
    return data


@register_compute_fun(
    name="g^zz_z",
    label="\\partial_{\\zeta} g^{\\zeta \\zeta}",
    units="m^-2",
    units_long="inverse square meters",
    description="Toroidal/Toroidal element of contravariant metric tensor, "
    + "first toroidal derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^zeta", "e^zeta_z"],
)
def _g_sup_zz_z(params, transforms, profiles, data, **kwargs):
    data["g^zz_z"] = 2 * dot(data["e^zeta_z"], data["e^zeta"])
    return data


@register_compute_fun(
    name="|grad(rho)|",
    label="|\\nabla \\rho|",
    units="m^{-1}",
    units_long="inverse meters",
    description="Magnitude of contravariant radial basis vector",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["g^rr"],
)
def _gradrho(params, transforms, profiles, data, **kwargs):
    data["|grad(rho)|"] = jnp.sqrt(data["g^rr"])
    return data


@register_compute_fun(
    name="<|grad(rho)|>",  # same as S(r) / V_r(r)
    label="\\langle \\vert \\nabla \\rho \\vert \\rangle",
    units="m^{-1}",
    units_long="inverse meters",
    description="Magnitude of contravariant radial basis vector, flux surface average",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["|grad(rho)|", "sqrt(g)"],
    axis_limit_data=["sqrt(g)_r"],
    resolution_requirement="tz",
)
def _gradrho_norm_fsa(params, transforms, profiles, data, **kwargs):
    data["<|grad(rho)|>"] = surface_averages(
        transforms["grid"],
        data["|grad(rho)|"],
        sqrt_g=transforms["grid"].replace_at_axis(
            data["sqrt(g)"], lambda: data["sqrt(g)_r"], copy=True
        ),
    )
    return data


@register_compute_fun(
    name="|grad(psi)|",
    label="|\\nabla\\psi|",
    units="Wb / m",
    units_long="Webers per meter",
    description="Toroidal flux gradient (normalized by 2pi) magnitude",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["|grad(psi)|^2"],
)
def _gradpsi_mag(params, transforms, profiles, data, **kwargs):
    data["|grad(psi)|"] = jnp.sqrt(data["|grad(psi)|^2"])
    return data


@register_compute_fun(
    name="|grad(psi)|^2",
    label="|\\nabla\\psi|^{2}",
    units="(Wb / m)^{2}",
    units_long="Webers squared per square meter",
    description="Toroidal flux gradient (normalized by 2pi) magnitude squared",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["grad(psi)"],
)
def _gradpsi_mag2(params, transforms, profiles, data, **kwargs):
    data["|grad(psi)|^2"] = dot(data["grad(psi)"], data["grad(psi)"])
    return data


@register_compute_fun(
    name="|grad(theta)|",
    label="|\\nabla \\theta|",
    units="m^{-1}",
    units_long="inverse meters",
    description="Magnitude of contravariant poloidal basis vector",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["g^tt"],
)
def _gradtheta(params, transforms, profiles, data, **kwargs):
    data["|grad(theta)|"] = jnp.sqrt(data["g^tt"])
    return data


@register_compute_fun(
    name="|grad(zeta)|",
    label="|\\nabla \\zeta|",
    units="m^{-1}",
    units_long="inverse meters",
    description="Magnitude of contravariant toroidal basis vector",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["g^zz"],
)
def _gradzeta(params, transforms, profiles, data, **kwargs):
    data["|grad(zeta)|"] = jnp.sqrt(data["g^zz"])
    return data


@register_compute_fun(
    name="g^aa",
    label="g^{\\alpha \\alpha}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Contravariant metric tensor grad alpha dot grad alpha",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["grad(alpha)"],
)
def _g_sup_aa(params, transforms, profiles, data, **kwargs):
    data["g^aa"] = dot(data["grad(alpha)"], data["grad(alpha)"])
    return data


@register_compute_fun(
    name="g^ra",
    label="g^{\\rho \\alpha}",
    units="m^{-2}",
    units_long="inverse square meters",
    description="Contravariant metric tensor grad rho dot grad alpha",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["grad(alpha)", "e^rho"],
)
def _g_sup_ra(params, transforms, profiles, data, **kwargs):
    data["g^ra"] = dot(data["grad(alpha)"], data["e^rho"])
    return data


@register_compute_fun(
    name="gbdrift",
    # Exact definition of the magnetic drifts taken from
    # eqn. 48 of Introduction to Quasisymmetry by Landreman
    # https://tinyurl.com/54udvaa4
    label="\\mathrm{gbdrift} = 1/B^{2} (\\mathbf{b}\\times\\nabla B) \\cdot"
    + "\\nabla \\alpha",
    units="1/(T-m^{2})",
    units_long="inverse Tesla meters^2",
    description="Binormal component of the geometric part of the gradB drift"
    + " used for local stability analyses, Gamma_c, epsilon_eff etc.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["|B|^2", "b", "grad(alpha)", "grad(|B|)"],
)
def _gbdrift(params, transforms, profiles, data, **kwargs):
    data["gbdrift"] = (
        1
        / data["|B|^2"]
        * dot(data["b"], cross(data["grad(|B|)"], data["grad(alpha)"]))
    )
    return data


@register_compute_fun(
    name="cvdrift",
    # Exact definition of the magnetic drifts taken from
    # eqn. 48 of Introduction to Quasisymmetry by Landreman
    # https://tinyurl.com/54udvaa4
    label="\\mathrm{cvdrift} = 1/B^{3} (\\mathbf{b}\\times\\nabla(p + B^2/2))"
    + "\\cdot \\nabla \\alpha",
    units="1/(T-m^{2})",
    units_long="inverse Tesla meters^2",
    description="Binormal component of the geometric part of the curvature drift"
    + " used for local stability analyses, Gamma_c, epsilon_eff etc.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["p_r", "psi_r", "|B|^2", "gbdrift"],
)
def _cvdrift(params, transforms, profiles, data, **kwargs):
    dp_dpsi = mu_0 * data["p_r"] / data["psi_r"]
    data["cvdrift"] = 1 / data["|B|^2"] * dp_dpsi + data["gbdrift"]
    return data


@register_compute_fun(
    name="cvdrift0",
    # Exact definition of the magnetic drifts taken from
    # eqn. 48 of Introduction to Quasisymmetry by Landreman
    # https://tinyurl.com/54udvaa4
    label="\\mathrm{cvdrift0} = 1/B^{2} (\\mathbf{b}\\times\\nabla B)"
    + "\\cdot \\nabla \\rho",
    units="1/(T-m^{2})",
    units_long="inverse Tesla meters^2",
    description="Radial component of the geometric part of the curvature drift"
    + " used for local stability analyses, gyrokinetics, Gamma_c.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["rho", "|B|^2", "b", "e^rho", "grad(|B|)"],
)
def _cvdrift0(params, transforms, profiles, data, **kwargs):
    data["cvdrift0"] = (
        2
        * data["rho"]
        / data["|B|^2"]
        * (dot(data["b"], cross(data["grad(|B|)"], data["e^rho"])))
    )
    return data



"""Tests for Mercier stability functions."""

import numpy as np
import pytest
from netCDF4 import Dataset
from scipy.interpolate import interp1d

import desc.examples
import desc.io
from desc.backend import jnp
from desc.equilibrium import Equilibrium
from desc.grid import Grid, LinearGrid, QuadratureGrid
from desc.objectives import MagneticWell, MercierStability
from desc.utils import PRINT_WIDTH, cross, dot

DEFAULT_RANGE = (0.05, 1)
DEFAULT_RTOL = 1e-2
DEFAULT_ATOL = 1e-6
MAX_SIGN_DIFF = 5


def assert_all_close(
    y1, y2, rho, rho_range=DEFAULT_RANGE, rtol=DEFAULT_RTOL, atol=DEFAULT_ATOL
):
    """Test that the values of y1 and y2, over a given range are close enough.

    Parameters
    ----------
    y1 : ndarray
        values to compare
    y2 : ndarray
        values to compare
    rho : ndarray
        rho values
    rho_range : (float, float)
        the range of rho values to compare
    rtol : float
        relative tolerance
    atol : float
        absolute tolerance

    """
    minimum, maximum = rho_range
    interval = (minimum < rho) & (rho < maximum)
    np.testing.assert_allclose(y1[interval], y2[interval], rtol=rtol, atol=atol)


def get_vmec_data(path, quantity):
    """Get data from a VMEC wout.nc file.

    Parameters
    ----------
    path : str
        Path to VMEC file.
    quantity: str
        Name of the quantity to return.

    Returns
    -------
    rho : ndarray
        Radial coordinate.
    q : ndarray
        Variable from VMEC output.

    """
    f = Dataset(path)
    phi = np.array(f.variables["phi"])
    rho = np.sqrt(phi / phi[-1])
    q = np.array(f.variables[quantity])
    f.close()
    return rho, q


@pytest.mark.unit
def test_mercier_vacuum():
    """Test that the Mercier stability criteria are 0 without pressure."""
    eq = Equilibrium()
    data = eq.compute(["D_shear", "D_current", "D_well", "D_geodesic", "D_Mercier"])
    np.testing.assert_allclose(data["D_shear"], 0)
    np.testing.assert_allclose(data["D_current"], 0)
    np.testing.assert_allclose(data["D_well"], 0)
    np.testing.assert_allclose(data["D_geodesic"], 0)
    np.testing.assert_allclose(data["D_Mercier"], 0)


@pytest.mark.unit
def test_compute_d_shear():
    """Test that D_shear has a stabilizing effect and matches VMEC."""

    def test(eq, vmec, rho_range=(0, 1), rtol=1e-12, atol=0.0):
        rho, d_shear_vmec = get_vmec_data(vmec, "DShear")
        grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, sym=eq.sym, rho=rho)
        d_shear = grid.compress(eq.compute("D_shear", grid=grid)["D_shear"])

        assert np.all(
            d_shear[bool(grid.axis.size) :] >= 0
        ), "D_shear should always have a stabilizing effect."
        assert_all_close(d_shear, d_shear_vmec, rho, rho_range, rtol, atol)

    test(
        desc.examples.get("DSHAPE_CURRENT"),
        ".//tests//inputs//wout_DSHAPE.nc",
        (0.3, 0.9),
        atol=0.01,
        rtol=0.1,
    )
    test(desc.examples.get("HELIOTRON"), ".//tests//inputs//wout_HELIOTRON.nc")


@pytest.mark.unit
def test_compute_d_current():
    """Test calculation of D_current stability criterion against VMEC."""

    def test(eq, vmec, rho_range=DEFAULT_RANGE, rtol=DEFAULT_RTOL, atol=DEFAULT_ATOL):
        rho, d_current_vmec = get_vmec_data(vmec, "DCurr")
        grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, sym=eq.sym, rho=rho)
        d_current = grid.compress(eq.compute("D_current", grid=grid)["D_current"])

        assert (
            np.nonzero(np.sign(d_current) != np.sign(d_current_vmec))[0].size
            <= MAX_SIGN_DIFF
        )
        assert_all_close(d_current, d_current_vmec, rho, rho_range, rtol, atol)

    test(
        desc.examples.get("DSHAPE_CURRENT"),
        ".//tests//inputs//wout_DSHAPE.nc",
        (0.3, 0.9),
        rtol=1e-1,
        atol=1e-2,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.25, 0.85),
        rtol=1e-1,
    )


@pytest.mark.unit
def test_compute_d_well():
    """Test calculation of D_well stability criterion against VMEC."""

    def test(eq, vmec, rho_range=DEFAULT_RANGE, rtol=DEFAULT_RTOL, atol=DEFAULT_ATOL):
        rho, d_well_vmec = get_vmec_data(vmec, "DWell")
        grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, sym=eq.sym, rho=rho)
        d_well = grid.compress(eq.compute("D_well", grid=grid)["D_well"])

        assert (
            np.nonzero(np.sign(d_well) != np.sign(d_well_vmec))[0].size <= MAX_SIGN_DIFF
        )
        assert_all_close(d_well, d_well_vmec, rho, rho_range, rtol, atol)

    test(
        desc.examples.get("DSHAPE_CURRENT"),
        ".//tests//inputs//wout_DSHAPE.nc",
        (0.3, 0.9),
        rtol=1e-1,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.01, 0.45),
        rtol=1.75e-1,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.45, 0.6),
        atol=7.2e-1,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.6, 0.99),
        rtol=2e-2,
    )


@pytest.mark.unit
def test_compute_d_geodesic():
    """Test that D_geodesic has a destabilizing effect and matches VMEC."""

    def test(eq, vmec, rho_range=DEFAULT_RANGE, rtol=DEFAULT_RTOL, atol=DEFAULT_ATOL):
        rho, d_geodesic_vmec = get_vmec_data(vmec, "DGeod")
        grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, sym=eq.sym, rho=rho)
        d_geodesic = grid.compress(eq.compute("D_geodesic", grid=grid)["D_geodesic"])

        assert np.all(
            d_geodesic[bool(grid.axis.size) :] <= 0
        ), "D_geodesic should always have a destabilizing effect."
        assert_all_close(d_geodesic, d_geodesic_vmec, rho, rho_range, rtol, atol)

    test(
        desc.examples.get("DSHAPE_CURRENT"),
        ".//tests//inputs//wout_DSHAPE.nc",
        (0.3, 0.9),
        rtol=1e-1,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.15, 0.825),
        rtol=1.2e-1,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.85, 0.95),
        atol=1.2e-1,
    )


@pytest.mark.unit
def test_compute_d_mercier():
    """Test calculation of D_Mercier stability criterion against VMEC."""

    def test(eq, vmec, rho_range=DEFAULT_RANGE, rtol=DEFAULT_RTOL, atol=DEFAULT_ATOL):
        rho, d_mercier_vmec = get_vmec_data(vmec, "DMerc")
        grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, sym=eq.sym, rho=rho)
        d_mercier = grid.compress(eq.compute("D_Mercier", grid=grid)["D_Mercier"])

        assert (
            np.nonzero(np.sign(d_mercier) != np.sign(d_mercier_vmec))[0].size
            <= MAX_SIGN_DIFF
        )
        assert_all_close(d_mercier, d_mercier_vmec, rho, rho_range, rtol, atol)

    test(
        desc.examples.get("DSHAPE_CURRENT"),
        ".//tests//inputs//wout_DSHAPE.nc",
        (0.3, 0.9),
        rtol=1e-1,
        atol=1e-2,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.1, 0.325),
        rtol=1.3e-1,
    )
    test(
        desc.examples.get("HELIOTRON"),
        ".//tests//inputs//wout_HELIOTRON.nc",
        (0.325, 0.95),
        rtol=5e-2,
    )


@pytest.mark.unit
def test_compute_magnetic_well():
    """Test that D_well and magnetic_well match signs under finite pressure."""

    def test(eq, rho=np.linspace(0, 1, 128)):
        grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, sym=eq.sym, rho=rho)
        d_well = grid.compress(eq.compute("D_well", grid=grid)["D_well"])
        magnetic_well = grid.compress(
            eq.compute("magnetic well", grid=grid)["magnetic well"]
        )
        assert (
            np.nonzero(np.sign(d_well) != np.sign(magnetic_well))[0].size
            <= MAX_SIGN_DIFF
        )

    test(desc.examples.get("DSHAPE_CURRENT"))
    test(desc.examples.get("HELIOTRON"))


@pytest.mark.unit
def test_mercier_print(capsys):
    """Test that the Mercier stability criteria prints correctly."""
    eq = Equilibrium()
    grid = LinearGrid(L=10, M=10, N=5, axis=False)

    Dmerc = eq.compute("D_Mercier", grid=grid)["D_Mercier"]

    mercier_obj = MercierStability(eq=eq, grid=grid)
    mercier_obj.build()
    np.testing.assert_allclose(mercier_obj.compute(*mercier_obj.xs(eq)), 0)
    mercier_obj.print_value(mercier_obj.xs(eq))
    out = capsys.readouterr()
    pre_width = len("Maximum ")

    corr_out = str(
        "Precomputing transforms\n"
        + "Maximum "
        + f"{mercier_obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.max(Dmerc))
        + mercier_obj._units
        + "\n"
        + "Minimum "
        + f"{mercier_obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.min(Dmerc))
        + mercier_obj._units
        + "\n"
        + "Average "
        + f"{mercier_obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.mean(Dmerc))
        + mercier_obj._units
        + "\n"
        + "Maximum "
        + f"{mercier_obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.max(Dmerc / mercier_obj.normalization))
        + "(normalized)"
        + "\n"
        + "Minimum "
        + f"{mercier_obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.min(Dmerc / mercier_obj.normalization))
        + "(normalized)"
        + "\n"
        + "Average "
        + f"{mercier_obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.mean(Dmerc / mercier_obj.normalization))
        + "(normalized)"
        + "\n"
    )
    assert out.out == corr_out


@pytest.mark.unit
def test_magwell_print(capsys):
    """Test that the magnetic well stability criteria prints correctly."""
    eq = desc.examples.get("HELIOTRON")
    grid = LinearGrid(L=12, M=12, N=6, NFP=eq.NFP, axis=False)
    obj = MagneticWell(eq=eq, grid=grid)
    obj.build()

    magwell = grid.compress(eq.compute("magnetic well", grid=grid)["magnetic well"])
    f = obj.compute(*obj.xs(eq))
    np.testing.assert_allclose(f, magwell)

    obj.print_value(obj.xs(eq))
    out = capsys.readouterr()
    pre_width = len("Maximum ")

    corr_out = str(
        "Precomputing transforms\n"
        + "Maximum "
        + f"{obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.max(magwell))
        + obj._units
        + "\n"
        + "Minimum "
        + f"{obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.min(magwell))
        + obj._units
        + "\n"
        + "Average "
        + f"{obj._print_value_fmt:<{PRINT_WIDTH - pre_width}}"
        + "{:10.3e} ".format(np.mean(magwell))
        + obj._units
        + "\n"
    )
    assert out.out == corr_out


@pytest.mark.unit
def test_ballooning_geometry(tmpdir_factory):
    """Test the geometry coefficients used for the adjoint-ballooning solver.

    We calculate coefficients using two different method and cross-compare them.
    The same coefficients are used for local gyrokinetic solvers which would
    be useful when we couple DESC with GX/GS2 etc.
    Observation: The larger the force error, the worse the tests behave. For
    example, HELIOTRON coefficients are hard to match
    """
    psi = 0.5  # Actually rho^2 (normalized)
    alpha = 0.0
    ntor = 2

    eq0 = desc.examples.get("W7-X")
    eq1 = desc.examples.get("precise_QA")

    eq_list = [eq0, eq1]

    for eq in eq_list:
        eq_keys = ["iota", "iota_r", "a", "rho", "psi"]

        grid = QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid, NFP=eq.NFP)
        data_eq = eq.compute(eq_keys, grid=grid)

        fi = interp1d(
            data_eq["rho"][grid.unique_rho_idx],
            data_eq["iota"][grid.unique_rho_idx],
        )
        fs = interp1d(
            data_eq["rho"][grid.unique_rho_idx],
            data_eq["iota_r"][grid.unique_rho_idx],
        )

        iotas = fi(np.sqrt(psi))
        shears = fs(np.sqrt(psi))

        rho = np.sqrt(psi)
        N = (8 * eq.M_grid * eq.N_grid) * ntor + 1
        zeta = np.linspace(-ntor * np.pi, ntor * np.pi, N)

        data_keys = [
            "p_r",
            "psi",
            "psi_r",
            "sqrt(g)_PEST",
            "|grad(psi)|^2",
            "grad(|B|)",
            "grad(alpha)",
            "grad(psi)",
            "B",
            "grad(|B|)",
            "kappa",
            "iota",
            "lambda_t",
            "lambda_z",
            "lambda_tt",
            "lambda_zz",
            "lambda_tz",
            "g^aa",
            "g^ra",
            "g^rr",
            "cvdrift",
            "cvdrift0",
            "|B|",
            "B^zeta",
            "theta_PEST_t",
            "g^tt",
            "g^zz",
            "g^rt",
            "g^tz",
            "g^rz",
        ]

        grid = Grid.create_meshgrid(
            [rho, alpha, zeta],
            coordinates="raz",
            period=(np.inf, 2 * np.pi, np.inf),
        )

        # Fieldline data
        data = eq.compute(data_keys, grid=grid)

        # Data used to defining normalization
        data01 = eq.compute(["Psi", "a"], grid=LinearGrid(rho=np.array([1.0])))
        # normalizations
        psi_b = data01["Psi"][-1] / (2 * jnp.pi)
        Lref = data01["a"]
        Bref = 2 * abs(psi_b) / Lref**2

        # Data used only used for signs
        psi_s = data_eq["psi"][-1]
        sign_psi = psi_s / np.abs(psi_s)
        sign_iota = iotas / np.abs(iotas)

        modB = data["|B|"]
        x = Lref * np.sqrt(psi)
        shat = -x / iotas * shears / Lref

        psi_r = data["psi_r"]
        grad_alpha = data["grad(alpha)"]

        g_sup_rr = data["g^rr"]
        g_sup_ra = data["g^ra"]
        g_sup_aa = data["g^aa"]

        lambda_r = data["lambda_r"]
        lambda_t = data["lambda_t"]
        lambda_z = data["lambda_z"]

        grad_alpha_r = lambda_r - zeta * shears
        grad_alpha_t = 1 + lambda_t
        grad_alpha_z = -iotas + lambda_z

        mod_grad_alpha_alt = np.sqrt(
            grad_alpha_r**2 * data["g^rr"]
            + grad_alpha_t**2 * data["g^tt"]
            + grad_alpha_z**2 * data["g^zz"]
            + 2 * grad_alpha_r * grad_alpha_t * data["g^rt"]
            + 2 * grad_alpha_r * grad_alpha_z * data["g^rz"]
            + 2 * grad_alpha_t * grad_alpha_z * data["g^tz"]
        )

        psi_r = 2 * psi_b * rho
        grad_psi_dot_grad_alpha_alt = (
            psi_r * grad_alpha_r * data["g^rr"]
            + psi_r * grad_alpha_t * data["g^rt"]
            + psi_r * grad_alpha_z * data["g^rz"]
        )

        grad_psi_dot_grad_psi_alt = psi_r**2 * data["g^rr"]

        modB = data["|B|"]
        B_sup_zeta = data["B^zeta"]

        gds2_alt = mod_grad_alpha_alt**2 * Lref**2 * psi
        gds2 = g_sup_aa * Lref**2 * psi

        gds21_alt = -sign_iota * grad_psi_dot_grad_alpha_alt * shat / Bref
        gds21 = -sign_iota * g_sup_ra * shat / Bref * (psi_r)

        gds22 = grad_psi_dot_grad_psi_alt * (1 / psi) * (shat / (Lref * Bref)) ** 2
        gds22_alt = g_sup_rr * (psi_r) ** 2 * (1 / psi) * (shat / (Lref * Bref)) ** 2

        gbdrift = np.array(dot(cross(data["B"], data["grad(|B|)"]), grad_alpha))
        gbdrift *= -sign_psi * 2 * Bref * Lref**2 / modB**3 * np.sqrt(psi)
        gbdrift_alt = -sign_psi * data["gbdrift"] * 2 * Bref * Lref**2 * np.sqrt(psi)

        cvdrift = (
            -sign_psi
            * 2
            * Bref
            * Lref**2
            * np.sqrt(psi)
            * dot(cross(data["B"], data["kappa"]), grad_alpha)
            / modB**2
        )
        cvdrift_alt = -sign_psi * data["cvdrift"] * 2 * Bref * Lref**2 * np.sqrt(psi)

        np.testing.assert_allclose(gds2, gds2_alt, rtol=4e-3)
        np.testing.assert_allclose(gds22, gds22_alt)
        # gds21 is a zero crossing quantity, rtol won't work,
        # shifting the grid slightly can change rtol requirement significantly
        np.testing.assert_allclose(gds21, gds21_alt, atol=1e-3)
        np.testing.assert_allclose(gbdrift, gbdrift_alt)
        # cvdrift is a zero crossing quantity, rtol won't work,
        # shifting the grid slightly can change rtol requirement significantly
        np.testing.assert_allclose(cvdrift, cvdrift_alt, atol=5e-3)

        sqrt_g_PEST = data["sqrt(g)_PEST"]
        np.testing.assert_allclose(sqrt_g_PEST, 1 / (B_sup_zeta / psi_r))


@pytest.mark.unit
def test_ballooning_stability_eval():
    """Cross-compare all the stability functions.

    We calculated the ideal ballooning growth rate and Newcomb ball
    metric for the HELIOTRON case at different radii.
    """
    from scipy.constants import mu_0

    eq = desc.examples.get("HELIOTRON")

    # Flux surfaces on which to evaluate ballooning stability
    surfaces = [0.01, 0.8, 1.0]

    grid = LinearGrid(rho=jnp.array(surfaces), NFP=eq.NFP)
    eq_data_keys = ["iota"]

    data = eq.compute(eq_data_keys, grid=grid)

    N_alpha = 8

    # Field lines on which to evaluate ballooning stability
    alpha = jnp.linspace(0, np.pi, N_alpha, endpoint=False)

    # Number of toroidal transits of the field line
    ntor = 3

    # Number of point along a field line in ballooning space
    N_zeta = 2 * ntor * eq.M_grid * eq.N_grid + 1

    # range of the ballooning coordinate zeta
    zeta = np.linspace(-jnp.pi * ntor, jnp.pi * ntor, N_zeta)

    for i in range(len(surfaces)):
        rho = np.array([surfaces[i]])

        grid = Grid.create_meshgrid(
            [rho, alpha, zeta],
            coordinates="raz",
            period=(np.inf, 2 * np.pi, np.inf),
        )

        data_keys0 = [
            "g^aa",
            "g^ra",
            "g^rr",
            "cvdrift",
            "cvdrift0",
            "|B|",
            "B^zeta",
            "p_r",
            "iota",
            "iota_r",
            "shear",
            "psi",
            "psi_r",
            "rho",
            "Psi",
        ]
        data0 = eq.compute(data_keys0, grid=grid)

        rho = jnp.mean(data0["rho"])

        data_keys01 = ["Psi", "a"]
        data01 = eq.compute(data_keys01, grid=LinearGrid(rho=np.array([1.0])))

        # here we use a different method for calculating the growth rate that uses
        # different numerics than "ideal ballooning lambda" so that we can verify them
        # against one another
        psi_b = data01["Psi"][-1] / (2 * jnp.pi)
        a_N = data01["a"]
        B_N = 2 * psi_b / a_N**2

        N_zeta0 = 15
        zeta0 = jnp.linspace(-0.5 * jnp.pi, 0.5 * jnp.pi, N_zeta0)

        iota = jnp.mean(data0["iota"])
        shear = jnp.mean(data0["shear"])
        psi = jnp.mean(data0["psi"])
        sign_psi = jnp.sign(psi)
        sign_iota = jnp.sign(iota)

        phi = zeta

        B = grid.meshgrid_reshape(data0["|B|"], "arz")
        B_sup_zeta = grid.meshgrid_reshape(data0["B^zeta"], "arz")
        gradpar = B_sup_zeta / B

        dpdpsi = jnp.mean(mu_0 * data0["p_r"] / data0["psi_r"])

        g_sup_aa = grid.meshgrid_reshape(data0["g^aa"], "arz")[None, ...]
        g_sup_ra = grid.meshgrid_reshape(data0["g^ra"], "arz")[None, ...]
        g_sup_rr = grid.meshgrid_reshape(data0["g^rr"], "arz")[None, ...]

        gds2 = jnp.reshape(
            rho**2
            * (
                g_sup_aa
                - 2 * sign_iota * shear / rho * zeta0[:, None, None, None] * g_sup_ra
                + zeta0[:, None, None, None] ** 2 * (shear / rho) ** 2 * g_sup_rr
            ),
            (N_alpha, N_zeta0, N_zeta),
        )

        f = a_N * B_N**3 * gds2 / B**3 * 1 / gradpar
        g = a_N**3 * B_N * gds2 / B * gradpar
        g_half = (g[:, :, 1:] + g[:, :, :-1]) / 2

        cvdrift = grid.meshgrid_reshape(data0["cvdrift"], "arz")[None, ...]
        cvdrift0 = grid.meshgrid_reshape(data0["cvdrift0"], "arz")[None, ...]

        c = (
            a_N**3
            * B_N
            * jnp.reshape(
                2
                / B_sup_zeta[None, ...]
                * sign_psi
                * rho**2
                * dpdpsi
                * (
                    cvdrift
                    - shear / (2 * rho**2) * zeta0[:, None, None, None] * cvdrift0
                ),
                (N_alpha, N_zeta0, N_zeta),
            )
        )

        h = phi[1] - phi[0]

        A = jnp.zeros((N_alpha, N_zeta0, N_zeta - 2, N_zeta - 2))

        i = jnp.arange(N_alpha)[:, None, None, None]
        l = jnp.arange(N_zeta0)[None, :, None, None]
        j = jnp.arange(N_zeta - 2)[None, None, :, None]
        k = jnp.arange(N_zeta - 2)[None, None, None, :]

        A = A.at[i, l, j, k].set(
            g_half[i, l, k] / f[i, l, k] / h**2 * (j - k == -1)
            + (
                -(g_half[i, l, j + 1] + g_half[i, l, j]) / f[i, l, j + 1] / h**2
                + c[i, l, j + 1] / f[i, l, j + 1]
            )
            * (j - k == 0)
            + g_half[i, l, j] / f[i, l, j + 1] / h**2 * (j - k == 1)
        )

        w = jnp.linalg.eigvals(jnp.where(jnp.isfinite(A), A, 0))

        lam1 = jnp.max(jnp.real(jnp.max(w, axis=(2,))))

        # now compute our regular metrics and compare them
        data_keys = ["ideal ballooning lambda", "Newcomb ballooning metric"]
        data = eq.compute(data_keys, grid=grid)

        lam2 = np.max(data["ideal ballooning lambda"])
        Newcomb_metric = data["Newcomb ballooning metric"]

        np.testing.assert_allclose(lam1, lam2, rtol=5e-5)

        if lam2 > 0:
            assert Newcomb_metric <= 0, (
                "Newcomb metric indicates stabiliy for an unstable equilibrium, "
                f"surface = {rho}, lam = {lam2}, newcomb = {Newcomb_metric}"
            )
        else:
            assert Newcomb_metric > 0, (
                "Newcomb metric indicates instabiliy for a stable equilibrium, "
                f"surface = {rho}, lam = {lam2}, newcomb = {Newcomb_metric}"
            )


@pytest.mark.unit
def test_ballooning_compare_with_COBRAVMEC():
    """Compare marginal stability points from DESC ballooning solve with COBRAVMEC.

    COBRAVMEC uses some higher order techniques to refine the growth rate value,
    so we don't expect the actual values to agree, but the sign should be the same.
    so instead of comparing growth rates, we compare the radial point where it
    becomes unstable. Recall that it should be stable on axis (zero pressure gradient)
    but can become unstable elsewhere.
    """

    def find_root_simple(x, y):
        sign_changes = np.where(np.diff(np.sign(y)))[0]

        if len(sign_changes) == 0:
            return None  # No zero crossing found

        # Get the indices where y changes sign
        i = sign_changes[0]

        # Linear interpolation
        x0, x1 = x[i], x[i + 1]
        y0, y1 = y[i], y[i + 1]

        # Calculate the zero crossing
        x_zero = x0 + (0 - y0) * (x1 - x0) / (y1 - y0)

        return x_zero

    A = np.loadtxt("./tests/inputs/cobra_grate.HELIOTRON_L24_M16_N12")

    ns1 = int(A[0, 2])
    nangles = int(np.shape(A)[0] / (ns1 + 1))

    B = np.zeros((ns1,))
    for i in range(nangles):
        if i == 0:
            B = A[i + 1 : (i + 1) * ns1 + 1, 2]
        else:
            B = np.vstack((B, A[i * ns1 + i + 1 : (i + 1) * ns1 + i + 1, 2]))

    gamma1 = np.amax(B, axis=0)

    s1 = np.linspace(0, 1, ns1)
    s1 = s1 + np.diff(s1)[0]

    # COBRAVMEC calculated everything in s(=rho^2),
    # DESC calculates in rho(=sqrt(s))
    rho1 = np.sqrt(s1)

    root_COBRAVMEC = find_root_simple(rho1, gamma1)

    eq = desc.examples.get("HELIOTRON")

    # Flux surfaces on which to evaluate ballooning stability
    surfaces = [0.98, 0.985, 0.99, 0.995, 1.0]

    grid = LinearGrid(rho=jnp.array(surfaces), NFP=eq.NFP)

    Nalpha = 8  # Number of field lines

    assert Nalpha == int(8), "Nalpha in the compute function hard-coded to 8!"

    # Field lines on which to evaluate ballooning stability
    alpha = jnp.linspace(0, np.pi, Nalpha + 1)[:Nalpha]

    # Number of toroidal transits of the field line
    ntor = 3

    # Number of point along a field line in ballooning space
    N0 = 2 * ntor * eq.M_grid * eq.N_grid + 1

    # range of the ballooning coordinate zeta
    zeta = np.linspace(-jnp.pi * ntor, jnp.pi * ntor, N0)

    lam2_array = np.zeros(
        len(surfaces),
    )
    for i in range(len(surfaces)):
        rho = surfaces[i]
        grid = Grid.create_meshgrid(
            [rho, alpha, zeta],
            coordinates="raz",
            period=(np.inf, 2 * np.pi, np.inf),
        )

        data_keys = ["ideal ballooning lambda"]
        data = eq.compute(data_keys, grid=grid)

        lam2_array[i] = np.max(data["ideal ballooning lambda"])

    root_DESC = find_root_simple(np.array(surfaces), lam2_array)
    np.testing.assert_allclose(root_COBRAVMEC, root_DESC, rtol=4e-4)
